-- ipbus_main
--
-- Top level of the ipbus bus master
--
-- Bridges ethernet MAC interface and / or out-of-band interfaces
-- to the ipbus.

-- for resend, and status destination IP address and port used are the last used
-- => no storage

-- for control packets IP address and port are saved in a memory buffer.
-- Hence multi controller are possible (if no reliability is required!)

-- Olivier Bourrion, Aug 2014

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.ipbus.all;
use work.pack_ipv4_types.all;

entity ipbus_main is 
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH --MAXIMUM is 14 bit
	ADDRWIDTH: natural := 9;
	-- UDP port for IPbus traffic in this instance of UDP I/F
	IPBUSPORT: std_logic_vector(15 DOWNTO 0) := x"C351"
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset
    
	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- ipb interface
	ipb_clk: in std_logic; -- IPbus clock
	rst_ipb: in std_logic; -- IPbus clock domain sync reset
	ipb_out: out ipb_wbus; -- IPbus bus signals
	ipb_in: in ipb_rbus
	);
end ipbus_main;

architecture RTL of ipbus_main is
component vect_domain_xfer
generic(size : positive);
port(
	clkin : in std_logic;
	rstin : in std_logic;
	vect_in : in std_logic_vector(size-1 downto 0);

	clkout : in std_logic;
	rstout : in std_logic;
	vect_out : out std_logic_vector(size-1 downto 0)
	);
end component;

component transactor_lpsc
generic(
	-- Number of RX and TX buffers is 2**BUFWIDTH
	BUFWIDTH: natural := 4;
	-- Number of address bits within each buffer in UDP I/F
	-- Size of each buffer is 2**ADDRWIDTH --MAXIMUM is 15 bit
	ADDRWIDTH: natural := 9
	);
port(
	ipb_clk: in std_logic; -- IPbus clock
	rst_ipb: in std_logic; -- IPbus clock domain sync reset

	--ipb bus
	ipb_out: out ipb_wbus; -- IPbus bus signals
	ipb_in: in ipb_rbus;

	-- status interface
	next_expected_packetid : out std_logic_vector(15 downto 0);

	-- rx buffer interface
	rxtrans_buf_wr_ptr : in std_logic_vector(BUFWIDTH downto 0);
	rxtrans_rd_ptr : out std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
	rxtrans_dat : in std_logic_vector(31 downto 0);

	-- tx buffer interface
	txtrans_buf_wr_ptr : out std_logic_vector(BUFWIDTH downto 0);
	txtrans_dat_wr_ptr : out std_logic_vector(ADDRWIDTH-1 downto 0);
	txtrans_we : out std_logic;
	txtrans_dat : out std_logic_vector(31 downto 0)
	);
end component;

component ipbus_dpram32
generic(ADDSIZE : natural:=13);
port(
	clk_in : in std_logic;
	wr_mem : in std_logic;
	add_in : in std_logic_vector(ADDSIZE-1 downto 0);
	data_in : in std_logic_vector(31 downto 0);

	clk_out : in std_logic;
	add_out : in std_logic_vector(ADDSIZE-1 downto 0);
	data_out : out std_logic_vector(31 downto 0)
	);
end component;


signal rx32_s : std_logic_vector(31 downto 0);
signal rx_cnt_s : natural range 0 to 3;
signal rx_wr32 : std_logic;
signal wr_rx_mem : std_logic;
signal inc_buf_ptr : std_logic;
signal in_rx_mem : std_logic_vector(31 downto 0);
signal hdr_valid_1ck : std_logic;
signal rx_error_1ck : std_logic;
signal rx_error_2ck : std_logic;
signal inc_status_req : std_logic;
signal dec_status_req : std_logic;
signal inc_resend_req : std_logic;
signal packet2resend : std_logic_vector(15 downto 0);
signal PC_src_port : std_logic_vector(15 downto 0);
signal PC_src_ip : std_logic_vector(31 downto 0);
signal status_port : std_logic_vector(15 downto 0);
signal status_ip : std_logic_vector(31 downto 0);
signal next_expected_packetid : std_logic_vector(15 downto 0);


type def_state_ipbus_rx is (IDLE,FIRST_RECEPTION,FINISH_RECEPTION,FINISH_RECEPTION2,REQUEST_STATUS,RESEND_PACKET);
signal state_ipbus_rx : def_state_ipbus_rx;

-- BUFFERS
signal txtrans_memout : std_logic_vector(31 downto 0);
signal status_mem : natural range 0 to (2**BUFWIDTH)-1;

type def_resend_mem is array (0 to (2**BUFWIDTH)-1) of std_logic_vector(15 downto 0);
signal resend_mem : def_resend_mem;
signal resend_wr_ptr : std_logic_vector(BUFWIDTH-1 downto 0);
signal resend_rd_ptr : std_logic_vector(BUFWIDTH-1 downto 0);
signal is_resending : std_logic;

signal rxtrans_wr_buf_ptr : std_logic_vector(BUFWIDTH downto 0);
signal rxtrans_wr_buf_ptr_ipb : std_logic_vector(BUFWIDTH downto 0);

signal trans_wr_dat_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal trans_wr_ptr : std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
signal is_LE : std_logic;
signal status_is_LE : std_logic;
signal rx32_count : std_logic_vector(ADDRWIDTH-1 downto 0);

signal rxtrans_rd_ptr : std_logic_vector(ADDRWIDTH+BUFWIDTH-1 downto 0);
signal rxtrans_dat : std_logic_vector(31 downto 0);
		
signal txtrans_buf_wr_ptr_ipb : std_logic_vector(BUFWIDTH downto 0);
signal txtrans_buf_wr_ptr : std_logic_vector(BUFWIDTH downto 0);
signal txtrans_dat_wr_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal txtrans_wr_ptr : std_logic_vector(BUFWIDTH+ADDRWIDTH-1 downto 0);
signal txtrans_we : std_logic;
signal txpacket_read : std_logic;
signal txtrans_dat_rd_ptr : std_logic_vector(ADDRWIDTH-1 downto 0);
signal txtrans_memout_ptr : std_logic_vector(BUFWIDTH+ADDRWIDTH-1 downto 0);
signal txtrans_dat : std_logic_vector(31 downto 0);
signal txtrans_memout_1ck : std_logic_vector(31 downto 0);
signal txtrans_dat_fwft : std_logic_vector(31 downto 0);

signal rd_next : std_logic;
signal rd_next_1ck : std_logic;
signal dsel : std_logic;
signal rd_init : std_logic;
signal rd_init_1ck : std_logic;

-- status packet signals
type def_packet_history is array (0 to 3) of std_logic_vector(31 downto 0);
signal rcvd_pkt_hdr : def_packet_history;
signal sent_pkt_hdr : def_packet_history;
type def_status_pkt is array (0 to 15) of std_logic_vector(31 downto 0);
signal status_pkt : def_status_pkt;

signal status_byte_sel : natural range 0 to 3;
signal status_word_sel : natural range 0 to 15;
signal status_byte_cnt : std_logic_vector(5 downto 0);
signal status_dataout_8 : std_logic_vector(7 downto 0);

type def_state_tx_status is (IDLE,WAIT_UDP,XMIT,FINISH);
signal state_tx_status : def_state_tx_status;

-- control packet signals (normal and resend)
signal txtrans_buf_rd_ptr : std_logic_vector(BUFWIDTH downto 0);
signal cur_txtrans_buf_rd_ptr : std_logic_vector(BUFWIDTH-1 downto 0);
signal tx_is_LE : std_logic;
signal tx_cnt_target : std_logic_vector(ADDRWIDTH-1 downto 0);
signal prev_tx_cnt_target : std_logic_vector(ADDRWIDTH-1 downto 0);
signal control_byte_cnt: std_logic_vector(1 downto 0);
signal control_byte_endianness: natural range 0 to 3;

type def_state_tx_control is (IDLE,DO_READ_INIT,PKT_PREPAR,
					BUF_LOOKUP,SET_LEN,SET_IP,WAIT_UDP,XMIT,FINISH);
signal state_tx_control : def_state_tx_control;
constant wait_fwft_val : natural:=3;
signal cpt_wait_fwft : natural range 0 to wait_fwft_val-1;

--arbitrator
constant nb_src : natural :=2;
signal req_udp : std_logic_vector(nb_src-1 downto 0);
signal grant_udp : std_logic_vector(nb_src-1 downto 0);
signal src_udp_tx_start : std_logic_vector(nb_src-1 downto 0);
signal src_udp_tx : udp_tx_vect_type(nb_src-1 downto 0);

type def_state_arb is (idle,busy);
signal state_arb_udp : def_state_arb;

signal sel : natural range 0 to nb_src-1;

begin
-------------------------------------------------------------------------------
-- Packet reception, transform to word 32
-------------------------------------------------------------------------------
proc_rx : process(mac_clk,rst_macclk)
begin
	if rst_macclk='1' then
		rx32_s<=(others=>'0');
		rx_cnt_s<=0;
		rx_wr32<='0';
		hdr_valid_1ck<='0';
		rx_error_1ck<='0';
	elsif rising_edge(mac_clk) then
		if udp_rxo.hdr.is_valid ='1' and udp_rxo.hdr.dst_port=IPBUSPORT then
			hdr_valid_1ck<='1';
			rx_error_1ck<=udp_rxo.hdr.rx_error;

			if udp_rxo.data.data_in_valid='1' then
				rx32_s<=udp_rxo.data.data_in & rx32_s(31 downto 8);
				if rx_cnt_s=3 then
					rx_wr32<='1';
					rx_cnt_s<=0;
				else
					rx_wr32<='0';
					rx_cnt_s<=rx_cnt_s+1;
				end if;
			end if;
		else
			hdr_valid_1ck<='0';
			rx_error_1ck<='0';
			rx_wr32<='0';
			rx_cnt_s<=0;
		end if;
	end if;
end process proc_rx;

process(mac_clk)
begin
	if rising_edge(mac_clk) then
		rx_error_2ck<=rx_error_1ck;
	end if;
end process;

-------------------------------------------------------------------------------
-- check header, can be control, status or resend
-------------------------------------------------------------------------------
rx_interpreter : process(mac_clk,rst_macclk)
begin
	if rst_macclk='1' then
		wr_rx_mem<='0';
		in_rx_mem<=(others=>'0');
		PC_src_port<=(others=>'0');
		PC_src_ip<=(others=>'0');
		wr_rx_mem<='0';
		inc_buf_ptr<='0';
		inc_status_req<='0';
		inc_resend_req<='0';
		packet2resend<=(others=>'0');
		rx32_count<=conv_std_logic_vector(1,ADDRWIDTH);
		state_ipbus_rx<=IDLE;
		rcvd_pkt_hdr<=(others=>(others=>'0'));
		is_LE<='0';
		trans_wr_dat_ptr<=conv_std_logic_vector(1,ADDRWIDTH);

	elsif rising_edge(mac_clk) then
		case state_ipbus_rx is
			when IDLE =>
				wr_rx_mem<='0';
				inc_buf_ptr<='0';
				rx32_count<=conv_std_logic_vector(1,ADDRWIDTH);
				PC_src_port<=udp_rxo.hdr.src_port;
				PC_src_ip<=udp_rxo.hdr.src_ip_addr;
				trans_wr_dat_ptr<=conv_std_logic_vector(1,ADDRWIDTH);
				if hdr_valid_1ck='1' then
					state_ipbus_rx<=FIRST_RECEPTION;
				end if;

			when FIRST_RECEPTION =>
				if rx_wr32='1' then
					if rx32_s(7 downto 4)=x"F" then
						is_LE<='1';
						case rx32_s(3 downto 0) is
							when x"0" => -- control
								rcvd_pkt_hdr(0)<=rx32_s;
								for i in 1 to 3 loop
									rcvd_pkt_hdr(i)<=rcvd_pkt_hdr(i-1);
								end loop;

								wr_rx_mem<='1';
								trans_wr_dat_ptr<=trans_wr_dat_ptr+1;
								in_rx_mem<=rx32_s;
								state_ipbus_rx<=FINISH_RECEPTION;
							when x"1" => -- status
								if rx_error_1ck='0' then
									inc_status_req<='1';
									state_ipbus_rx<=REQUEST_STATUS;
								end if;
							when x"2" => -- resend
								if rx_error_1ck='0' then
									inc_resend_req<='1';
									packet2resend<=rx32_s(23 downto 8);
									state_ipbus_rx<=RESEND_PACKET;
								end if;
							when others => --unknown
						end case;
					else
						is_LE<='0';
						case rx32_s(27 downto 24) is
							when x"0" => -- control
								rcvd_pkt_hdr(0)<=rx32_s(7 downto 0) &rx32_s(15 downto 8) &rx32_s(23 downto 16) & rx32_s(31 downto 24);
								for i in 1 to 3 loop
									rcvd_pkt_hdr(i)<=rcvd_pkt_hdr(i-1);
								end loop;
								wr_rx_mem<='1';
								trans_wr_dat_ptr<=trans_wr_dat_ptr+1;
								in_rx_mem<=rx32_s;
								state_ipbus_rx<=FINISH_RECEPTION;
							when x"1" => -- status
								if rx_error_1ck='0' then
									inc_status_req<='1';
									state_ipbus_rx<=REQUEST_STATUS;
								end if;
							when x"2" => -- resend
								if rx_error_1ck='0' then
									inc_resend_req<='1';
									packet2resend<=rx32_s(15 downto 8) & rx32_s(23 downto 16);
									state_ipbus_rx<=RESEND_PACKET;
								end if;
							when others => --unknown
						end case;
					end if;
				else
					wr_rx_mem<='0';
				end if;
			
			when FINISH_RECEPTION =>
				if hdr_valid_1ck='1' then
					if rx_wr32='1' then
						rx32_count<=rx32_count+1; -- size of ipbus packet + 1 (for 1st word)
						wr_rx_mem<='1';
						trans_wr_dat_ptr<=trans_wr_dat_ptr+1;
						in_rx_mem<=rx32_s;
					else
						wr_rx_mem<='0';
					end if;
				else
					wr_rx_mem<='1';
					trans_wr_dat_ptr<=conv_std_logic_vector(0,ADDRWIDTH);
					in_rx_mem(31 downto 16)<=PC_src_port;
					in_rx_mem(15)<=is_LE;
					in_rx_mem(14 downto 0)<=EXT(rx32_count,15);
					state_ipbus_rx<=FINISH_RECEPTION2;
				end if;
			
			when FINISH_RECEPTION2 =>
					if rx_error_2ck='0' then
						inc_buf_ptr<='1';
					end if;
					wr_rx_mem<='1';
					trans_wr_dat_ptr<=conv_std_logic_vector(1,ADDRWIDTH);
					in_rx_mem(31 downto 0)<=PC_src_ip;
					state_ipbus_rx<=IDLE;

			when REQUEST_STATUS => -- no checking of the content
				inc_status_req<='0';
				if hdr_valid_1ck='0' then
					state_ipbus_rx<=IDLE;
				end if;

			when RESEND_PACKET =>
				inc_resend_req<='0';
				if hdr_valid_1ck='0' then
					state_ipbus_rx<=IDLE;
				end if;
		end case;
	end if;
end process rx_interpreter;
trans_wr_ptr<=rxtrans_wr_buf_ptr(BUFWIDTH-1 downto 0) & trans_wr_dat_ptr;
-------------------------------------------------------------------------------
-- rx buffer manager
-------------------------------------------------------------------------------
lbl_rx_transmem: ipbus_dpram32 generic map(ADDSIZE=>ADDRWIDTH+BUFWIDTH)
port map(
	clk_in=>mac_clk,
	wr_mem=>wr_rx_mem,
	add_in=>trans_wr_ptr,
	data_in=>in_rx_mem,

	clk_out=>ipb_clk,
	add_out=>rxtrans_rd_ptr,
	data_out=>rxtrans_dat
	);

buff_mgr : process(mac_clk,rst_macclk)
begin
	if rst_macclk='1' then
		status_mem<=0;
		status_is_LE<='0';
		resend_wr_ptr<=(others=>'0');
		rxtrans_wr_buf_ptr<=(others=>'0');
		status_port<=(others=>'0');
		status_ip<=(others=>'0');

	elsif rising_edge(mac_clk) then
		-- status request FIFO
		if (inc_status_req='1' and dec_status_req='0') then
			status_is_LE<=is_LE;
			status_port<=PC_src_port;
			status_ip<=PC_src_ip;
			if status_mem/=(2**BUFWIDTH)-1 then
				status_mem<=status_mem+1;
			end if;
		elsif (inc_status_req='0' and dec_status_req='1') then
			if status_mem/=0 then
				status_mem<=status_mem-1;
			end if;
		end if;

		-- packet resend request FIFO
		if inc_resend_req='1' then
			resend_mem(conv_integer(resend_wr_ptr))<=packet2resend;
			resend_wr_ptr<=resend_wr_ptr+1;
		end if;
		
		-- transaction buffer write
		if inc_buf_ptr='1' then
			rxtrans_wr_buf_ptr<=rxtrans_wr_buf_ptr+1;
		end if;
	end if;
end process buff_mgr;
-------------------------------------------------------------------------------
-- transfer to ipb_clk and to transactor_lpsc
-------------------------------------------------------------------------------
lbl_to_ipb : vect_domain_xfer generic map(size=>BUFWIDTH+1)
	port map(
		clkin=>mac_clk,
		rstin=>rst_macclk,
		vect_in=>rxtrans_wr_buf_ptr,

		clkout=>ipb_clk,
		rstout=>rst_ipb,
		vect_out=>rxtrans_wr_buf_ptr_ipb
		);
-------------------------------------------------------------------------------
-- transactor_lpsc
-------------------------------------------------------------------------------
lbl_transactor_lpsc : transactor_lpsc generic map(BUFWIDTH=>BUFWIDTH,ADDRWIDTH=>ADDRWIDTH)
	port map(
		ipb_clk=>ipb_clk,
		rst_ipb=>rst_ipb,

		--ipb bus
		ipb_out=>ipb_out,
		ipb_in=>ipb_in,

		-- rx buffer interface
		rxtrans_buf_wr_ptr=>rxtrans_wr_buf_ptr_ipb,
		rxtrans_rd_ptr=>rxtrans_rd_ptr,
		rxtrans_dat=>rxtrans_dat,

		-- status interface
		next_expected_packetid=>next_expected_packetid, -- !!!generated on ipb_clk!!!

		-- tx buffer interface
		txtrans_buf_wr_ptr=>txtrans_buf_wr_ptr_ipb,
		txtrans_dat_wr_ptr=>txtrans_dat_wr_ptr,
		txtrans_we=>txtrans_we,
		txtrans_dat=>txtrans_dat
		);
-------------------------------------------------------------------------------
-- tx transaction memory interface
-------------------------------------------------------------------------------
lbl_tx_transmem: ipbus_dpram32 generic map(ADDSIZE=>ADDRWIDTH+BUFWIDTH)
port map(
	clk_in=>ipb_clk,
	wr_mem=>txtrans_we,
	add_in=>txtrans_wr_ptr,
	data_in=>txtrans_dat,

	clk_out=>mac_clk,
	add_out=>txtrans_memout_ptr,
	data_out=>txtrans_memout
	);
txtrans_wr_ptr<=txtrans_buf_wr_ptr_ipb(BUFWIDTH-1 downto 0) & txtrans_dat_wr_ptr;
-------------------------------------------------------------------------------
-- transfer to mac_clk and to transmitter
-------------------------------------------------------------------------------
lbl_to_mac : vect_domain_xfer generic map(size=>BUFWIDTH+1)
	port map(
		clkin=>ipb_clk,
		rstin=>rst_ipb,
		vect_in=>txtrans_buf_wr_ptr_ipb,

		clkout=>mac_clk,
		rstout=>rst_macclk,
		vect_out=>txtrans_buf_wr_ptr
		);
-------------------------------------------------------------------------------
-- status packet
-------------------------------------------------------------------------------
status_pkt(0)<=x"20_0000_F1";
status_pkt(1)<=conv_std_logic_vector(2**(ADDRWIDTH+2)-8,32);
status_pkt(2)<=conv_std_logic_vector(2**BUFWIDTH,32);
status_pkt(3)<=x"20" & next_expected_packetid & x"F0"; -- !!!generated on ipb_clk!!!
status_pkt(4)<=(others=>'0'); --traffic history not yet supported
status_pkt(5)<=(others=>'0'); --traffic history not yet supported
status_pkt(6)<=(others=>'0'); --traffic history not yet supported
status_pkt(7)<=(others=>'0'); --traffic history not yet supported
status_pkt(8)<=rcvd_pkt_hdr(0);
status_pkt(9)<=rcvd_pkt_hdr(1);
status_pkt(10)<=rcvd_pkt_hdr(2);
status_pkt(11)<=rcvd_pkt_hdr(3);
status_pkt(12)<=sent_pkt_hdr(0);
status_pkt(13)<=sent_pkt_hdr(1);
status_pkt(14)<=sent_pkt_hdr(2);
status_pkt(15)<=sent_pkt_hdr(3);

-------------------------------------------------------------------------------
-- status transmitter FSM
-------------------------------------------------------------------------------
src_udp_tx(1).hdr.src_port<=IPBUSPORT;
src_udp_tx(1).hdr.data_length<=conv_std_logic_vector(64,16);
src_udp_tx(1).hdr.checksum<=(others=>'0');
-- most recent inbound transaction is used for filling this fields
-- makes senses since it is used ofr reliability and this works only with one client

-- not yet fully reliable (if several source for status request, the latest 
-- requested will be served if 2 short consecutives requests
src_udp_tx(1).hdr.dst_ip_addr<=status_ip; 
src_udp_tx(1).hdr.dst_port<=status_port;

status_word_sel<=conv_integer(status_byte_cnt(5 downto 2));
status_byte_sel<=	conv_integer(status_byte_cnt(1 downto 0)) when status_is_LE='1' else
					conv_integer(not(status_byte_cnt(1 downto 0)));

with status_byte_sel select
status_dataout_8<=	status_pkt(status_word_sel)(7 downto 0) when 0,
					status_pkt(status_word_sel)(15 downto 8) when 1,
					status_pkt(status_word_sel)(23 downto 16) when 2,
					status_pkt(status_word_sel)(31 downto 24) when 3;

proc_tx_status : process(mac_clk,rst_macclk)
begin
	if rst_macclk='1' then
		req_udp(1)<='0';
		dec_status_req<='0';
	
		src_udp_tx_start(1)<='0';
		
		src_udp_tx(1).data.data_out_valid<='0';
		src_udp_tx(1).data.data_out_last<='0';
		src_udp_tx(1).data.data_out<=(others=>'0');
		status_byte_cnt<=(others=>'0');
		state_tx_status<=IDLE;

	elsif rising_edge(mac_clk) then
		case state_tx_status is
			when IDLE =>
				status_byte_cnt<=(others=>'0');
				dec_status_req<='0';
				if status_mem/=0 then
					req_udp(1)<='1';
					state_tx_status<=WAIT_UDP;
				end if;

			when WAIT_UDP =>
				if grant_udp(1)='1' then
					src_udp_tx_start(1)<='1';
					src_udp_tx(1).data.data_out_valid<='1';
					src_udp_tx(1).data.data_out<=status_dataout_8;
					status_byte_cnt<=status_byte_cnt+1;
					state_tx_status<=XMIT;
				end if;

			when XMIT =>
				if udp_tx_data_out_ready='1' then
					src_udp_tx(1).data.data_out<=status_dataout_8;
					if status_byte_cnt=63 then
						dec_status_req<='1';
						state_tx_status<=FINISH;
						src_udp_tx(1).data.data_out_last<='1';
					else
						status_byte_cnt<=status_byte_cnt+1;
					end if;
				end if;

			when FINISH =>
				if udp_tx_data_out_ready='1' then
					src_udp_tx(1).data.data_out_valid<='0';
					src_udp_tx(1).data.data_out_last<='0';
					src_udp_tx(1).data.data_out<=(others=>'0');
					req_udp(1)<='0';
					src_udp_tx_start(1)<='0';
					state_tx_status<=IDLE;
				end if;
		end case;
	end if;
end process proc_tx_status;
-------------------------------------------------------------------------------
-- fwft ram
-------------------------------------------------------------------------------
fwft : process(rst_macclk,mac_clk)
begin
	if rst_macclk='1' then
		txtrans_dat_rd_ptr<=(others=>'0');
		txtrans_memout_1ck<=(others=>'0');
		rd_init_1ck<='0';
		rd_next_1ck<='0';
		txpacket_read<='0';
	elsif rising_edge(mac_clk) then
		if rd_init='1' then
			txtrans_dat_rd_ptr<=(others=>'0');
		elsif rd_next='1' or rd_init_1ck='1' then
			txtrans_dat_rd_ptr<=txtrans_dat_rd_ptr+1;
		end if;
		
		rd_init_1ck<=rd_init;
		rd_next_1ck<=rd_next;
		dsel <= rd_next and rd_next_1ck;
			
		if rd_init = '1' or rd_init_1ck = '1' or rd_next = '1' or rd_next_1ck = '1' then
			txtrans_memout_1ck<=txtrans_memout;
		end if;
		
		if rd_init='1' then
			txpacket_read<='0';
		elsif rd_next = '1' then
			if txtrans_dat_rd_ptr=tx_cnt_target then
				txpacket_read<='1';
			end if;
		end if;
			
	end if;
end process fwft;

txtrans_dat_fwft <= txtrans_memout_1ck when dsel = '0' else txtrans_memout;
txtrans_memout_ptr<= cur_txtrans_buf_rd_ptr & txtrans_dat_rd_ptr;
-------------------------------------------------------------------------------
-- control transmitter FSM
-------------------------------------------------------------------------------
src_udp_tx(0).hdr.src_port<=IPBUSPORT;
src_udp_tx(0).hdr.checksum<=(others=>'0');

control_byte_endianness<=	conv_integer(control_byte_cnt) when tx_is_LE='1' else
							conv_integer(not(control_byte_cnt));

proc_tx_control : process(mac_clk,rst_macclk)
begin
	if rst_macclk='1' then
		req_udp(0)<='0';
		sent_pkt_hdr<=(others=>(others=>'0'));
	
		src_udp_tx_start(0)<='0';
		
		src_udp_tx(0).data.data_out_valid<='0';
		src_udp_tx(0).data.data_out_last<='0';
		src_udp_tx(0).data.data_out<=(others=>'0');
		src_udp_tx(0).hdr.data_length<=(others=>'0');
	
		tx_cnt_target<=(others=>'1');
		prev_tx_cnt_target<=(others=>'1');

		txtrans_buf_rd_ptr<=(others=>'0');
		cur_txtrans_buf_rd_ptr<=(others=>'0');
		tx_is_LE<='0';

		resend_rd_ptr<=(others=>'0');
		control_byte_cnt<=(others=>'0');
		is_resending<='0';
		state_tx_control<=IDLE;
		cpt_wait_fwft<=0;
		rd_init<='0';
		rd_next<='0';

	elsif rising_edge(mac_clk) then
		case state_tx_control is
			when IDLE =>
				tx_cnt_target<=(others=>'1');
				rd_init<='1';
				rd_next<='0';
				control_byte_cnt<=(others=>'0');

				if resend_rd_ptr/=resend_wr_ptr then
					cur_txtrans_buf_rd_ptr<=(others=>'0');
					is_resending<='1';
					state_tx_control<=DO_READ_INIT;
				elsif txtrans_buf_wr_ptr/=txtrans_buf_rd_ptr then
					is_resending<='0';
					cur_txtrans_buf_rd_ptr<=txtrans_buf_rd_ptr(BUFWIDTH-1 downto 0);
					state_tx_control<=DO_READ_INIT;
				else
					is_resending<='0';
				end if;

			when DO_READ_INIT =>
				cpt_wait_fwft<=0;
				state_tx_control<=PKT_PREPAR;

			when PKT_PREPAR =>
				rd_init<='0';
				if cpt_wait_fwft=wait_fwft_val-1 then
-- 					cpt_wait_fwft<=0;
					rd_next<='1';
					state_tx_control<=SET_IP;
				else
					cpt_wait_fwft<=cpt_wait_fwft+1;
				end if;

			when SET_IP =>
				rd_next<='1';
				src_udp_tx(0).hdr.dst_ip_addr<=txtrans_dat_fwft(31 downto 0);
				state_tx_control<=SET_LEN;

			when SET_LEN =>
				rd_next<='0';
				src_udp_tx(0).hdr.data_length<=txtrans_dat_fwft(13 downto 0)& "00"; -- conv length32 in bytes
				src_udp_tx(0).hdr.dst_port<=txtrans_dat_fwft(31 downto 16);
				prev_tx_cnt_target<=txtrans_dat_fwft(ADDRWIDTH-1 downto 0);
				tx_is_LE<=txtrans_dat_fwft(15);

				if is_resending='0' then
					req_udp(0)<='1';
					state_tx_control<=WAIT_UDP;
				else
					state_tx_control<=BUF_LOOKUP;
				end if;

			when BUF_LOOKUP =>
				if (txtrans_memout(23 downto 8)=resend_mem(conv_integer(resend_rd_ptr)) and prev_tx_cnt_target/=0) then -- compare to memory direct output to get header at add=2 and check valid content (size/=0)
					req_udp(0)<='1';
					resend_rd_ptr<=resend_rd_ptr+1;
					state_tx_control<=WAIT_UDP;

				else
					if cur_txtrans_buf_rd_ptr=2**BUFWIDTH-1 then -- not found!
						resend_rd_ptr<=resend_rd_ptr+1;
						state_tx_control<=IDLE;
					else
						cur_txtrans_buf_rd_ptr<=cur_txtrans_buf_rd_ptr+1;
						rd_init<='1';
						state_tx_control<=DO_READ_INIT;
					end if;
				end if;
			

			when WAIT_UDP =>
				if grant_udp(0)='1' then
					tx_cnt_target<=prev_tx_cnt_target+1; -- to cope with added IP add (length+1)
					src_udp_tx_start(0)<='1';
					src_udp_tx(0).data.data_out_valid<='1';
					
					case control_byte_endianness is
						when 0 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(7 downto 0);
						when 1 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(15 downto 8);
						when 2 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(23 downto 16);
						when 3 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(31 downto 24);
					end case;

					sent_pkt_hdr(0)<=txtrans_dat_fwft;
					for i in 1 to 3 loop
						sent_pkt_hdr(i)<=sent_pkt_hdr(i-1);
					end loop;

					control_byte_cnt<=control_byte_cnt+1;
					state_tx_control<=XMIT;
				end if;

			when XMIT =>
				if udp_tx_data_out_ready='1' then
					case control_byte_endianness is
						when 0 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(7 downto 0);
						when 1 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(15 downto 8);
						when 2 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(23 downto 16);
						when 3 => src_udp_tx(0).data.data_out<=txtrans_dat_fwft(31 downto 24);
					end case;
					control_byte_cnt<=control_byte_cnt+1;

					if control_byte_cnt=2 then
						rd_next<='1';
					else
						rd_next<='0';
					end if;

					if control_byte_cnt=3 then
						if txpacket_read='1' then
							state_tx_control<=FINISH;
							src_udp_tx(0).data.data_out_last<='1';
						end if;
					end if;

				elsif udp_tx_result = UDPTX_RESULT_ERR then
					src_udp_tx(0).data.data_out_valid<='0';
					src_udp_tx(0).data.data_out_last<='0';
					src_udp_tx(0).data.data_out<=(others=>'0');
					req_udp(0)<='0';
					src_udp_tx_start(0)<='0';
					if is_resending='0' then
						txtrans_buf_rd_ptr<=txtrans_buf_rd_ptr+1;
					end if;
					state_tx_control<=IDLE;
				end if;

			when FINISH =>
				if udp_tx_data_out_ready='1' then
					src_udp_tx(0).data.data_out_valid<='0';
					src_udp_tx(0).data.data_out_last<='0';
					src_udp_tx(0).data.data_out<=(others=>'0');
					req_udp(0)<='0';
					src_udp_tx_start(0)<='0';
					if is_resending='0' then
						txtrans_buf_rd_ptr<=txtrans_buf_rd_ptr+1;
					end if;
					state_tx_control<=IDLE;
				end if;
		end case;
	end if;
end process proc_tx_control;
-------------------------------------------------------------------------------
-- UDP arbitrer
-------------------------------------------------------------------------------
process(mac_clk)
begin
	if rising_edge(mac_clk) then
		if rst_macclk = '1' then
			sel<=0;
			for i in 0 to nb_src-1 loop
				grant_udp(i)<='0';
			end loop;
			state_arb_udp<=idle;
		else
			case state_arb_udp is
				when idle =>
					if req_udp(sel)='1' then
						grant_udp(sel)<='1';
						state_arb_udp<=busy;
					elsif sel=nb_src-1 then
						sel<=0;
					else
						sel<=sel+1;
					end if;

				when busy =>
					if src_udp_tx(sel).data.data_out_last='1' then
						grant_udp(sel)<='0';
						sel<=0;
						state_arb_udp<=idle;
					end if;
			end case;
		end if;
	end if;
end process;

-- multiplexer
udp_tx_start <= src_udp_tx_start(sel);
udp_txi <= src_udp_tx(sel);

end RTL;
