--------------------------------------------------
-- ntp_local_time
-- 
-- uses ntp_sub_add, hence i:
--		a minimum of 7 refclk must be garanted between 2 sub_sec_top to not miss any tick
--		a minimum of of 4 refclk if missing a clock tick during update is allowed
--
-- It is expected to be several second/min/hours between update_thetae, 
-- hence theta stay valid for a long time, no need to register it
-- 
-- O. Bourrion May 2015
--------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ntp_local_time is 
generic(
   increment_bit: natural := 20 -- conputed form DDS log_P ->increment_bit=32-log_P
	);
port(
	refclk: in std_logic; -- Stable reference clock
	rst: in std_logic; -- rst
    
	sub_sec_top: in std_logic;
	update_theta: in std_logic;
	theta : in std_logic_vector(63 downto 0);

	new_time_avail: out std_logic; -- signal to use for clock domain synchronisation
	local_time : out std_logic_vector(63 downto 0)
	);
end ntp_local_time;

architecture RTL of ntp_local_time is
component ntp_addsub is 
generic(
	is_adder : boolean :=true;
	size : natural :=64 -- must be even number here
	);
port(
	clk : in std_logic;

	op_1 : in std_logic_vector(size-1 downto 0);
	op_2 : in std_logic_vector(size-1 downto 0);

	result : out std_logic_vector(size-1 downto 0)
	);
end component ntp_addsub;

signal time_increment : std_logic_vector(63 downto 0);
signal internal_time : std_logic_vector(63 downto 0);
signal s_local_time : std_logic_vector(63 downto 0);
signal saved_update_theta : std_logic;
signal theta_done : std_logic;

signal update_cnt : natural range 0 to 2;

type def_state_time_mgr is (idle, time_incr, time_theta);
signal state_time_mgr : def_state_time_mgr;

begin
-------------------------------------------------------------------------------
-- time counter -- two refclk cycles to update!
-------------------------------------------------------------------------------
lbl_time : ntp_addsub generic map(
	is_adder=>true,
	size=>64
	)
port map(
	clk=>refclk,

	op_1=>internal_time,
	op_2=>time_increment,

	result=>s_local_time
	);
local_time<=internal_time;
-------------------------------------------------------------------------------
-- ntp adder
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if theta_done='1' or rst='1' then
			saved_update_theta<='0';
		elsif update_theta='1' then
			saved_update_theta<='1';
		end if;

		if (rst='1') then
			state_time_mgr<=idle;
			time_increment<=(others=>'0');
			internal_time<=(others=>'0');
			theta_done<='0';
			new_time_avail<='0';
		else
			case state_time_mgr is
				when idle =>
					update_cnt<=0;
					theta_done<='0';
					new_time_avail<='0';
					if sub_sec_top='1' then
-- 						time_increment<=(increment_bit=>'1',others=>'0'); -- +2**increment_bit
						for i in 0 to time_increment'length-1 loop
							if i=increment_bit then
								time_increment(i)<='1'; -- +2**increment_bit
							else
								time_increment(i)<='0';
							end if;
						end loop;
						state_time_mgr<=time_incr;
					end if;
				
				when time_incr =>
					if update_cnt=2 then
						update_cnt<=0;
						internal_time<=s_local_time;
						if saved_update_theta='1' then
							time_increment<=theta;
							state_time_mgr<=time_theta; -- + theta
						else
							new_time_avail<='1';
							state_time_mgr<=idle;
						end if;
					else
						update_cnt<=update_cnt+1;
						time_increment<=(others=>'0');
					end if;

				when time_theta =>
					if update_cnt=2 then
						update_cnt<=0;
						theta_done<='1';
						internal_time<=s_local_time;
						new_time_avail<='1';
						state_time_mgr<=idle;
					else
						update_cnt<=update_cnt+1;
						time_increment<=(others=>'0');
					end if;

			end case;
		end if;
	end if;
end process;
end RTL;
