----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 		Peter Fall, rewrite by Olivier Bourrion (fusion of process, sync between tx and RX)
-- 
-- Create Date:    12:00:04 05/31/2011 
-- Design Name: 
-- Module Name:    arp - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description:
--		handle simple IP lookup in cache
--		request cache fill through ARP protocol if required
--		cache is simple 1 deep
--		Handle ARP protocol
--		Respond to ARP requests and replies
--		Ignore pkts that are not ARP
--		Ignore pkts that are not addressed to us
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Revision 0.02 - Added req for mac tx and wait for grant
-- Revision 0.03 - Added data_out_first
-- Revision 0.04 - Added arp response timeout
-- Revision 0.05 - Added arp cache reset control
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.pack_arp_types.all;

entity arp is
	 generic (
			IN_SIMULATION     : boolean :=false;
			SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
			CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
			ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
			ARP_MAX_PKT_TMO	: integer := 1;									-- (added for compatibility with arpv2. this value not used in this impl)
			MAX_ARP_ENTRIES 	: integer := 1										-- (added for compatibility with arpv2. this value not used in this impl)
			);
    Port (
			-- lookup request signals
			arp_req_req			: in arp_req_req_type;
			arp_req_rslt		: out arp_req_rslt_type;
			-- MAC layer RX signals
			data_in_clk 		: in  STD_LOGIC;
			reset 				: in  STD_LOGIC;
			data_in 				: in  STD_LOGIC_VECTOR (7 downto 0);		-- ethernet frame (from dst mac addr through to last byte of frame)
			data_in_valid 		: in  STD_LOGIC;									-- indicates data_in valid on clock
			data_in_error        : in  std_logic;							-- tells mac produced an error
			data_in_last 		: in  STD_LOGIC;									-- indicates last data in frame
			-- MAC layer TX signals
			mac_tx_req			: out std_logic;									-- indicates that ip wants access to channel (stays up for as long as tx)
			mac_tx_granted		: in std_logic;									-- indicates that access to channel has been granted		
			data_out_clk		: in std_logic;
			data_out_ready		: in std_logic;									-- indicates system ready to consume data
			data_out_valid		: out std_logic;									-- indicates data out is valid
			data_out_first		: out std_logic;									-- with data out valid indicates the first byte of a frame
			data_out_last		: out std_logic;									-- with data out valid indicates the last byte of a frame
			data_out				: out std_logic_vector (7 downto 0);		-- ethernet frame (from dst mac addr through to last byte of frame)
			-- system signals
			our_mac_address 	: in STD_LOGIC_VECTOR (47 downto 0);
			our_ip_address 	: in STD_LOGIC_VECTOR (31 downto 0);
			control				: in arp_control_type;
			req_count			: out STD_LOGIC_VECTOR(7 downto 0)			-- count of arp pkts received
			);
end arp;

architecture Behavioral of arp is
component arp_store
generic (
    MAX_ARP_ENTRIES : integer := 16 -- max entries in the store
    );
port (
    -- system signals
    clk         : in  std_logic;
    reset       : in  std_logic;

    -- arp lookup signals
    req_ip    : in  std_logic_vector(31 downto 0);
	do_lookup : in std_logic;
	lookup_done : out std_logic;
    rslt_mac    : out  std_logic_vector(47 downto 0);
	req_found : out std_logic;

	-- arp table write signals
    wr_ip    : in  std_logic_vector(31 downto 0);
    wr_mac    : in  std_logic_vector(47 downto 0);
	insert_arp : in std_logic;
	insert_done : out std_logic;

    -- control and status signals
    clear_store : in  std_logic;              -- erase all entries
    entry_count : out std_logic_vector(7 downto 0)   -- how many entries currently in store
    );
end component;

type req_state_type is (IDLE,CACHE_LOOKUP,REQUEST,WAIT_REPLY,PAUSE1,PAUSE2,PAUSE3);
-- type rx_state_type is (IDLE,PARSE,PROCESS_ARP,WAIT_END);
type rx_state_type is (IDLE,PARSE,WAIT_END);
type rx_event_type is (NO_EVENT,DATA);
type count_mode_type is (RST,INCR,HOLD);
type arp_oper_type is (NOP,REQUEST,REPLY);
type set_clr_type is (SET, CLR, HOLD);

type tx_state_type is (IDLE,WAIT_MAC,SEND);


type arp_entry_type is record
	ip 				: std_logic_vector (31 downto 0);
	mac 				: std_logic_vector (47 downto 0);
	is_valid			: std_logic;
	reply_required	: std_logic;
end record;

-- state variables
signal req_state			: req_state_type;
signal send_request_needed		: std_logic;
signal freq_scaler		: unsigned (31 downto 0);		-- scales data_in_clk downto 1Hz
signal timer				: unsigned (7 downto 0);		-- counts seconds timeout

signal rx_count 			: unsigned (7 downto 0);
signal arp_operation		: arp_oper_type;
signal arp_req_count		: unsigned (7 downto 0);
signal arp_entry			: arp_entry_type;				-- arp entry store
signal new_arp_entry		: arp_entry_type;
signal tx_state			: tx_state_type;
signal tx_count 			: unsigned (7 downto 0);

-- requester control signals
signal timer_enable		: std_logic;				-- enable the timer counting

signal reply_required_1ck	: std_logic;
signal reply_required_2ck	: std_logic;

signal dst_ip : std_logic_vector(31 downto 0);
signal dst_mac : std_logic_vector(47 downto 0);
signal rslt_mac : std_logic_vector(47 downto 0);

-- rx control signals
signal rx_state 	: rx_state_type;
signal set_arp_entry_request		: std_logic;

-- tx control signals
signal clear_reply_req	: std_logic;
signal clear_reply_req_1ck	: std_logic;
signal clear_reply_req_2ck	: std_logic;

signal do_lookup : std_logic;
signal insert_done : std_logic;
signal lookup_done : std_logic;
signal req_found : std_logic;
-- function to determine whether the rx pkt is an arp pkt and whether we want to process it
-- Returns 1 if we should discard
-- The following will make us ignore the frame (all values hexadecimal):
-- PDU type /= 0806
-- Protocol Type /= 0800
-- Hardware Type /= 1
-- Hardware Length /= 6
-- Protocol Length /= 4
-- Operation /= 1 or 2
-- Target IP /= our IP (i.er. message is not meant for us)
--
function not_our_arp(data : STD_LOGIC_VECTOR; count : unsigned; our_ip : std_logic_vector) return std_logic is
begin
	if 
		(count = 12 and data /= x"08") or						-- PDU type 0806 : ARP
		(count = 13 and data /= x"06") or
		(count = 14 and data /= x"00") or						-- HW type 1 : eth
		(count = 15 and data /= x"01") or
		(count = 16 and data /= x"08") or						-- Protocol 0800 : IP
		(count = 17 and data /= x"00") or
		(count = 18 and data /= x"06") or						-- HW Length 6
		(count = 19 and data /= x"04") or						-- protocol length 4
		(count = 20 and data /= x"00") or						-- operation 1 or 2 (req or reply)
		(count = 21 and data /= x"01" and data /= x"02") or
		(count = 38 and data /= our_ip(31 downto 24)) or	-- target IP is ours
		(count = 39 and data /= our_ip(23 downto 16)) or
		(count = 40 and data /= our_ip(15 downto 8)) or
		(count = 41 and data /= our_ip(7 downto 0))	
	then
		return '1';
	else
		return '0';
	end if;
end function not_our_arp;
begin
-------------------------------------------------------------------------------
-- ARP STORE INST
-------------------------------------------------------------------------------
lbl_arp_store :  arp_store generic map(MAX_ARP_ENTRIES=>MAX_ARP_ENTRIES)
port map(
    -- system signals
    clk=>data_in_clk,
    reset=>reset,

    -- arp lookup signals
    req_ip=>arp_req_req.ip,
	do_lookup=>do_lookup,
	lookup_done=>lookup_done,
    rslt_mac=>rslt_mac,
	req_found=>req_found,

	-- arp table write signals
    wr_ip=>new_arp_entry.ip,
    wr_mac=>new_arp_entry.mac,
	insert_arp=>set_arp_entry_request,
	insert_done=>insert_done,

    -- control and status signals
    clear_store=>control.clear_cache,
    entry_count=>open
    );


-------------------------------------------------------------------------------
-- REQ FSM
-------------------------------------------------------------------------------
request_process : process (data_in_clk)
begin
	if rising_edge(data_in_clk) then
		if reset = '1' then
			freq_scaler <= to_unsigned(CLOCK_FREQ,32);
			req_state<=idle;
			arp_req_rslt.mac<= (others => '0');
			arp_req_rslt.got_mac <= '0';
			send_request_needed <= '0';
			do_lookup<='0';
		else
			-- freq scaling and 1-sec timer
			if freq_scaler = x"00000000" then
				freq_scaler <= to_unsigned(CLOCK_FREQ,32);
				timer_enable<='1';
			else
				freq_scaler <= freq_scaler - 1;
				timer_enable<='0';
			end if;
				
			-- REQ FSM
			case req_state is
				when IDLE =>
					timer <= x"00";
					arp_req_rslt.got_mac <= '0';
					if arp_req_req.lookup_req = '1' then
						do_lookup<='1';
						arp_req_rslt.got_err <= '0';
						req_state <= CACHE_LOOKUP;
					else
						send_request_needed <= '0';
					end if;

				when CACHE_LOOKUP =>
					do_lookup<='0';
					if (lookup_done='1') then
-- 						if (req_found='1' or IN_SIMULATION) then -- already have this IP
-- 							if IN_SIMULATION then
						if (req_found='1') then -- already have this IP
							arp_req_rslt.mac<= rslt_mac;
							arp_req_rslt.got_mac <= '1';
							req_state <= IDLE;
						else
							req_state <= REQUEST;
							timer <= x"00";
							arp_req_rslt.mac<= (others => '0');
							arp_req_rslt.got_mac <= '0';
						end if;
					end if;

				when REQUEST =>
						send_request_needed <= '1';
						req_state <= WAIT_REPLY;
					
				when WAIT_REPLY =>
						send_request_needed <= '0';
						if insert_done = '1' then -- store received a reply, go back to CACHE_LOOKUP state to see if it is the right one
							do_lookup<='1';
							req_state <= CACHE_LOOKUP;
						end if;
						if timer_enable = '1' then
							timer <= timer + 1;
						end if;

						if timer >= ARP_TIMEOUT then
							arp_req_rslt.got_err <= '1';
							req_state <= PAUSE1;
						end if;

				when PAUSE1 =>
						req_state <= PAUSE2;

				when PAUSE2 =>
						req_state <= PAUSE3;

				when PAUSE3 =>
						req_state <= IDLE;
			end case;		
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- RX FSM
-------------------------------------------------------------------------------
rx_sequential : process (data_in_clk)
begin
	if rising_edge(data_in_clk) then
		if reset = '1' then
			-- reset state variables
			rx_state <= IDLE;
			rx_count <= x"00";
			arp_req_count <= x"00";
			arp_entry.reply_required <= '0';
			set_arp_entry_request <= '0';
		else
			clear_reply_req_1ck<=clear_reply_req;
			clear_reply_req_2ck<=clear_reply_req_1ck;

			case rx_state is
				when IDLE =>
					arp_operation <= NOP;
					set_arp_entry_request <= '0';
					if data_in_valid = '1' then
						rx_state <= PARSE;
						rx_count <= rx_count + 1;
					else
						rx_count <= x"00";
					end if;

				when PARSE =>
					if data_in_valid = '1' then
							rx_count <= rx_count + 1;
							-- handle early frame termination
							if data_in_last = '1' then
								rx_state <= IDLE;
							else
								-- check for end of frame. Also, detect and discard if not our frame
								if rx_count = 42 then
									rx_state <= WAIT_END;
								elsif not_our_arp(data_in,rx_count,our_ip_address) = '1' then
									arp_operation <= NOP;
									rx_state <= WAIT_END;
								elsif rx_count = 21 then
									-- capture ARP operation
									case data_in is
										when x"01" =>
											arp_operation <= REQUEST;
										when x"02" =>
											arp_operation <= REPLY;
										when others =>	-- ignore other values
									end case;
								-- capture source mac addr
								elsif rx_count = 22 then
									new_arp_entry.mac(47 downto 40) <= data_in;
								elsif rx_count = 23 then
									new_arp_entry.mac(39 downto 32) <= data_in;
								elsif rx_count = 24 then
									new_arp_entry.mac(31 downto 24) <= data_in;
								elsif rx_count = 25 then
									new_arp_entry.mac(23 downto 16) <= data_in;
								elsif rx_count = 26 then
									new_arp_entry.mac(15 downto 8) <= data_in;
								elsif rx_count = 27 then
									new_arp_entry.mac(7 downto 0) <= data_in;
								-- capture source ip addr
								elsif rx_count = 28 then
									new_arp_entry.ip(31 downto 24) <= data_in;
								elsif rx_count = 29 then
									new_arp_entry.ip(23 downto 16) <= data_in;
								elsif rx_count = 30 then
									new_arp_entry.ip(15 downto 8) <= data_in;
								elsif rx_count = 31 then
									new_arp_entry.ip(7 downto 0) <= data_in;
								end if;							
							end if;							
					else
						rx_count <= x"00";
					end if;

-- 				when PROCESS_ARP =>
-- 					rx_state <= WAIT_END;

				when WAIT_END =>
					if data_in_valid = '1' then
						if data_in_last = '1' then
							if data_in_error='0' then
								case arp_operation is
									when NOP => -- (nothing to do)
									when REQUEST =>
											set_arp_entry_request <= '1';
									when REPLY =>
											set_arp_entry_request <= '1';
								end case;

							end if;
							rx_state <= IDLE;
						end if;
					end if;
			end case;
			
			-----------------
			-- ARP reply cache control	
			-----------------
			if control.clear_cache = '1' then
				arp_entry.ip <= x"00000000";
				arp_entry.mac <= x"000000000000";
				arp_entry.is_valid <= '0';
				arp_entry.reply_required <= '0';
			
			elsif set_arp_entry_request = '1' then
				-- copy info from new entry to arp_entry and set reply required
				arp_entry.mac <= new_arp_entry.mac;
				arp_entry.ip <= new_arp_entry.ip;
				arp_entry.is_valid <= '1';
				if arp_operation = REQUEST then
					arp_entry.reply_required <= '1';
				else
					arp_entry.reply_required <= '0';
				end if;
				-- count another ARP pkt received
				arp_req_count <= arp_req_count + 1;
			
			elsif clear_reply_req_2ck = '1' then
				-- note: clear_reply_req is set by tx logic, but handled in the clk domain of the rx
				-- maintain arp entry state, but reset the reply required flag
				arp_entry.reply_required <= '0';
			
			elsif send_request_needed = '1' then
				-- set up the arp entry to take the request to be transmitted out by the TX FSM
				arp_entry.mac <= (others => '0');
				arp_entry.is_valid <= '0';
				arp_entry.reply_required <= '0';
			end if;
		end if;
	end if;
end process;
req_count <= STD_LOGIC_VECTOR(arp_req_count);
-------------------------------------------------------------------------------
-- TX FSM
-------------------------------------------------------------------------------
tx_sequential : process (data_out_clk)
begin
	if rising_edge(data_out_clk) then
		if reset = '1' then
			tx_state <= IDLE;
			mac_tx_req <= '0';	
			data_out_valid <= '0';
			data_out_first <= '0';
			data_out_last <= '0';
			data_out <= x"00";
			clear_reply_req <= '0';
			
			reply_required_1ck<='0';
			reply_required_2ck<='0';
		else
			reply_required_1ck<=arp_entry.reply_required;
			reply_required_2ck<=reply_required_1ck;
			-- TX FSM
			case tx_state is
				when IDLE =>
					tx_count <= x"00";
					clear_reply_req <= '0';
					if  reply_required_2ck= '1' or send_request_needed = '1' then
						mac_tx_req <= '1';	
						tx_state <= WAIT_MAC;
					end if;

				when WAIT_MAC =>
					if mac_tx_granted = '1' then
						data_out_valid <= '1';
						data_out_first <= '1';
						if reply_required_2ck = '1' then
							data_out <= arp_entry.mac (47 downto 40);
						else
							data_out <= x"ff";			-- dst = broadcast
						end if;
						tx_count <= tx_count + 1;
						tx_state <= SEND;
					end if;
					-- TODO - should handle timeout here
						
				when SEND =>
					if data_out_ready = '1' then
						data_out_first <= '0';
						tx_count <= tx_count + 1;

						case tx_count is
							when x"01"  => 
								if reply_required_2ck = '1' then
									data_out <= arp_entry.mac (39 downto 32);
								else
									data_out <= x"ff";			-- dst = broadcast
								end if;
							when x"02"  => 
								if reply_required_2ck = '1' then
									data_out <= arp_entry.mac (31 downto 24);
								else
									data_out <= x"ff";			-- dst = broadcast
								end if;
							when x"03"  => 
								if reply_required_2ck = '1' then
									data_out <= arp_entry.mac (23 downto 16);
								else
									data_out <= x"ff";			-- dst = broadcast
								end if;
							when x"04"  => 
								if reply_required_2ck = '1' then
									data_out <= arp_entry.mac (15 downto 8);
								else
									data_out <= x"ff";			-- dst = broadcast
								end if;
							when x"05"  => 
								if reply_required_2ck = '1' then
									data_out <= arp_entry.mac (7 downto 0);
								else
									data_out <= x"ff";			-- dst = broadcast
								end if;
							when x"06"  => data_out <= our_mac_address (47 downto 40); 	-- src = our mac
							when x"07"  => data_out <= our_mac_address (39 downto 32); 
							when x"08"  => data_out <= our_mac_address (31 downto 24); 
							when x"09"  => data_out <= our_mac_address (23 downto 16); 
							when x"0a"  => data_out <= our_mac_address (15 downto 8); 
							when x"0b"  => data_out <= our_mac_address (7 downto 0); 
							when x"0c"  => data_out <= x"08"; 									-- pkt type = 0806 : ARP
							when x"0d"  => data_out <= x"06"; 
							when x"0e"  => data_out <= x"00"; 									-- HW type = 0001 : eth
							when x"0f"  => data_out <= x"01"; 
							when x"10"  => data_out <= x"08"; 									-- protocol = 0800 : ip
							when x"11"  => data_out <= x"00"; 
							when x"12"  => data_out <= x"06"; 									-- HW size = 06
							when x"13"  => data_out <= x"04"; 									-- prot size = 04
							when x"14"  =>	data_out <= x"00"; 									-- opcode =		
							when x"15"  =>
								clear_reply_req <= '1';			-- reset the reply request (done in the rx clk process domain)
								if reply_required_2ck = '1' then
									dst_mac<=arp_entry.mac;
									dst_ip<=arp_entry.ip;
									data_out <= x"02";																			-- 02 : REPLY if arp_entry valid
								else
									dst_mac<=(others=>'0');
									dst_ip<=arp_req_req.ip;
									data_out <= x"01";																			-- 01 : REQ if arp_entry invalid
								end if;
							when x"16" => data_out <= our_mac_address (47 downto 40); 	-- sender mac
							when x"17" => data_out <= our_mac_address (39 downto 32); 
							when x"18" => data_out <= our_mac_address (31 downto 24); 
							when x"19" => data_out <= our_mac_address (23 downto 16); 
							when x"1a" => data_out <= our_mac_address (15 downto 8); 
							when x"1b" => data_out <= our_mac_address (7 downto 0); 
							when x"1c" => data_out <= our_ip_address (31 downto 24); 	-- sender ip
							when x"1d" => data_out <= our_ip_address (23 downto 16); 
							when x"1e" => data_out <= our_ip_address (15 downto 8); 
							when x"1f" => data_out <= our_ip_address (7 downto 0); 
							when x"20" => data_out <= dst_mac (47 downto 40); 	-- target mac 
							when x"21" => data_out <= dst_mac (39 downto 32); 
							when x"22" => data_out <= dst_mac (31 downto 24); 
							when x"23" => data_out <= dst_mac (23 downto 16); 
							when x"24" => data_out <= dst_mac (15 downto 8); 
							when x"25" => data_out <= dst_mac (7 downto 0); 
							when x"26" => data_out <= dst_ip  (31 downto 24); 	-- target ip
							when x"27" => data_out <= dst_ip   (23 downto 16); 
							when x"28" => data_out <= dst_ip   (15 downto 8);
							
							when x"29" =>
								data_out <= dst_ip(7 downto 0); 
								data_out_last <= '1';
							
							when x"2a" =>
-- 								data_out <= (others=>'0');
								data_out_last <= '0';
								data_out_valid <= '0';
								mac_tx_req <= '0';	
								tx_state <= IDLE;

							when others =>
								tx_state <= IDLE;
						end case;
					end if;
				end case;
			end if;
		end if;
end process;
end Behavioral;

