----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 		Peter Fall, rewrite by Olivier Bourrion (fusion of process, sync between tx and RX)
-- 
-- Create Date:    5 June 2011 
-- Design Name: 
-- Module Name:    UDP_RX - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--              handle simple UDP RX
--              doesnt check the checsum
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Revision 0.02 - Improved error handling
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use work.pack_axi.all;
use work.pack_ipv4_types.all;

entity ping is
  port (
    -- system signals
    clk          : in  std_logic;
    reset        : in  std_logic;
    our_ip_address     : in  std_logic_vector (31 downto 0);
    -- IP layer RX signals
    ip_rx_start  : in  std_logic;       -- indicates receipt of ip header
    ip_rx        : in  ipv4_rx_type;
	
	-- arbitrator signals
	ip_layer_request : out std_logic;
	ip_layer_granted : in std_logic;

	-- IP layer TX signals
    clk_out			        : in  std_logic;
	ip_tx_start				: out std_logic;
	ip_tx					: out ipv4_tx_type;							-- IP tx cxns
	ip_tx_result			: in std_logic_vector (1 downto 0);		-- tx status (changes during transmission)
	ip_tx_data_out_ready	: in std_logic									-- indicates IP TX is ready to take data
    );                  
end ping;

architecture Behavioral of ping is
type rx_state_ping_type is (IDLE, ICMP_HDR, USER_DATA, WAIT_END);
type crc_state_type is (IDLE, IDENT,USER_DATA,ADDOVF,FINAL);
type tx_state_ping_type is (IDLE,REQUEST,PAUSE, SEND_ICMP_HDR, SEND_USER_DATA);

type def_mem is array (0 to 2047) of std_logic_vector(7 downto 0);

-- state variables
signal rx_state_ping         : rx_state_ping_type;
signal rx_count         : unsigned (7 downto 0);
signal identifier         : std_logic_vector (15 downto 0);
signal seq_num         : std_logic_vector (15 downto 0); 
signal total_length      : std_logic_vector (15 downto 0); 
signal data16      : std_logic_vector (15 downto 0); 
signal do_crc : std_logic;
signal do_reply : std_logic;
signal do_reply_1ck : std_logic;
signal do_reply_2ck : std_logic;
signal checksum         : unsigned (23 downto 0); 

signal mem : def_mem;
signal wr_ptr : unsigned(10 downto 0):=(others=>'0'); -- does not require to be synthesized
signal rd_ptr : unsigned(10 downto 0);
signal data_cnt : unsigned(10 downto 0); 

signal crc_state: crc_state_type;
signal tx_state_ping         : tx_state_ping_type;
signal tx_count         : unsigned (3 downto 0);
signal s_ip_layer_request : std_logic;

begin
-------------------------------------------------------------------------------
-- ICMP receive
-------------------------------------------------------------------------------
rx_sequential : process (clk)
begin
    if rising_edge(clk) then
		if reset = '1' then
			-- reset state variables
			rx_state_ping         <= IDLE;
			rx_count         <= x"00";
			do_crc<='0';
		else
			total_length<="00000" & std_logic_vector(8+wr_ptr);

			case rx_state_ping is
				when IDLE =>
					do_reply<='0';
					do_crc<='0';
					if ip_rx.data.data_in_valid = '1' then
						wr_ptr<=(others=>'0');
						if ip_rx.hdr.protocol = x"01" then -- ICMP protocol
							rx_count <= rx_count + 1; 
							rx_state_ping <= ICMP_HDR;
						else -- non-ICMP protocol - ignore this pkt
							rx_count <= x"00";
							rx_state_ping <= WAIT_END;
						end if;
					end if;

				when ICMP_HDR =>
					if ip_rx.data.data_in_valid = '1' then
						if rx_count = x"07" then
							rx_count<= x"00";
							rx_state_ping <= USER_DATA;
						else
							rx_count <= rx_count + 1; 
						end if;
												
						if ip_rx.data.data_in_last = '1' then -- handle early frame termination
							rx_state_ping <= IDLE;
						else
							case rx_count is
								when x"00" => 
									if (ip_rx.data.data_in/=x"08") then -- not echo request
										rx_state_ping <= WAIT_END;
									end if;
								when x"01" => 
									if (ip_rx.data.data_in/=x"00") then -- not echo request
										rx_state_ping <= WAIT_END;
									end if;
								
								when x"02" => NULL; -- ignore checksum
								when x"03" => NULL; -- ignore checksum
								when x"04" =>	identifier(15 downto 8)<=ip_rx.data.data_in;
								when x"05" =>	identifier(7 downto 0)<=ip_rx.data.data_in;
								when x"06" =>	
									seq_num(15 downto 8)<=ip_rx.data.data_in;
									do_crc<='1'; -- must start here to phase chk computation with data
								when x"07" =>	
									seq_num(7 downto 0)<=ip_rx.data.data_in;
								when others =>  -- ignore other bytes in icmp header                                                                              
							end case;
						end if;
					end if;
				
				when USER_DATA =>
					if ip_rx.data.data_in_valid = '1' then
						wr_ptr<=wr_ptr+1;
						data16<=data16(7 downto 0) & ip_rx.data.data_in;
						mem(to_integer(wr_ptr))<=ip_rx.data.data_in;
						do_crc<='0';
						
						if ip_rx.data.data_in_last = '1' then
							if ip_rx.hdr.last_error_code= RX_EC_NONE then -- reply if no CRC error
								do_reply<='1';
							end if;
							rx_state_ping    <= IDLE;
						end if;
					end if;

				when WAIT_END =>
					if ip_rx.data.data_in_valid = '1' then
						if ip_rx.data.data_in_last = '1' then
						  rx_state_ping <= IDLE;
						end if;
					end if;
			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- compute checksum
-------------------------------------------------------------------------------
crc : process (clk_out)
begin
	if rising_edge(clk_out) then
		if (reset='1') then
				crc_state  <= IDLE;
		else
			case crc_state is
				when IDLE =>
					if do_crc = '1' then
						checksum<=x"00" & unsigned(identifier);
-- 						checksum<=x"00" & unsigned(identifier)+x"000800"; -- for test and comparaison with request
						crc_state  <= IDENT;
					end if;
				when IDENT =>
					checksum<=unsigned(checksum)+unsigned(seq_num);
					crc_state  <= USER_DATA;
				
				when USER_DATA =>
					if (wr_ptr(0)='0') then -- every even number update checksum
						checksum<=unsigned(checksum)+unsigned(data16);
					end if;
					if (do_reply='1') then
						crc_state  <= ADDOVF;
					end if;

				when ADDOVF => -- 16 LSB + carry
					checksum <= (unsigned(checksum) and x"00ffff")+ unsigned(checksum(23 downto 16));
					crc_state  <= FINAL;

				when FINAL => -- flip every bit
					checksum <= not(checksum);
					crc_state  <= IDLE;
			end case;
		end if;
	end if;
end process crc;
-------------------------------------------------------------------------------
-- ICMP send (echo reply)
-------------------------------------------------------------------------------
ip_tx.hdr.dst_ip_addr <= ip_rx.hdr.src_ip_addr;
ip_tx.hdr.protocol <= x"01";	-- ICMP protocol
ip_tx.hdr.data_length <= total_length;
ip_layer_request<=s_ip_layer_request;

proc_icmp_send : process (clk_out)
begin
    if rising_edge(clk_out) then
		if reset = '1' then
			-- reset state variables
			tx_state_ping         <= IDLE;
			tx_count         <= x"0";
			ip_tx_start <='0';
			ip_tx.data.data_out_last <= '0';
			ip_tx.data.data_out_valid <='0';
			s_ip_layer_request <='0';
			do_reply_1ck<='0';
			do_reply_2ck<='0';
		else
			do_reply_1ck<=do_reply;
			do_reply_2ck<=do_reply_1ck;

			case tx_state_ping is
				when IDLE =>
					data_cnt<=wr_ptr;
					ip_tx.data.data_out_last <= '0';
					ip_tx.data.data_out_valid <='0';
					rd_ptr<=(others=>'0');
					tx_count <= x"0";
					if do_reply_2ck='1' then
						s_ip_layer_request <='1';
						tx_state_ping <= REQUEST;
					end if;

				when REQUEST =>
					if ip_layer_granted='1' then
						ip_tx_start <= '1';
						tx_state_ping <= PAUSE;
					end if;

				when PAUSE =>
					-- delay one clock for IP layer to respond to ip_tx_start and remove any tx error result
					ip_tx.data.data_out_valid <='1';
					tx_count<= tx_count+1;
					ip_tx.data.data_out <= x"00"; -- type
					tx_state_ping <= SEND_ICMP_HDR;

				when SEND_ICMP_HDR =>
					if ip_tx_result = IPTX_RESULT_ERR then
						ip_tx_start <= '0';
						ip_tx.data.data_out_valid <='0';
						tx_state_ping <= IDLE;
					elsif ip_tx_data_out_ready = '1' then
						if tx_count = x"7" then
							if data_cnt/=0 then
								data_cnt<=data_cnt-1;
								tx_state_ping <= SEND_USER_DATA;
							else
								ip_tx.data.data_out_last <= '1';
								s_ip_layer_request <='0';
								tx_state_ping <= IDLE;
							end if;
						else
							tx_count<= tx_count+1;
						end if;
						case tx_count is
							when x"1"  => ip_tx.data.data_out <= x"00"; -- code
							when x"2"  => ip_tx.data.data_out <= std_logic_vector(checksum(15 downto 8));
							when x"3"  => ip_tx.data.data_out <= std_logic_vector(checksum(7 downto 0));
							when x"4"  => ip_tx.data.data_out <= identifier(15 downto 8);
							when x"5"  => ip_tx.data.data_out <= identifier(7 downto 0);
							when x"6"  => ip_tx.data.data_out <= seq_num(15 downto 8);
							when x"7"  => ip_tx.data.data_out <= seq_num(7 downto 0);
							when others =>
								NULL;
						end case;
					end if;
					
				when SEND_USER_DATA =>
					ip_tx.data.data_out<= mem(to_integer(rd_ptr));
					if ip_tx_data_out_ready = '1' then
						rd_ptr<=rd_ptr+1; -- rd_ptr +1
						if rd_ptr = data_cnt then						
							-- TX terminated due to count - end normally
							ip_tx.data.data_out_last <= '1';
							ip_tx_start <= '0';
							s_ip_layer_request <='0';
							tx_state_ping <= IDLE;
						end if;
					end if;
				end case;
			end if;
		end if;
end process proc_icmp_send;

end Behavioral;

